##############################################################################################
## Start with potential cloud shadow mask

#### Band 5 flood fill
nir=raster("im_B5.tif")

# estimating background (land) Band 5 ref
backg_B5<-quantile(mask(nir,idlnd,maskvalue=1),l_pt,na.rm=T)
nir[mask[]==0]=backg_B5

# fill in regional minimum Band 5 ref
in.fill.NIR <- gsub(".grd$", ".sgrd", rasterTmpFile())      
writeRaster(nir,in.fill.NIR,format="SAGA",overwrite=TRUE,datatype = "FLT4S")
rsaga.fill.sinks(in.fill.NIR, in.fill.NIR, method = "xxl.wang.liu.2006", cores=1)
out.filled.NIR <- gsub(".sgrd$", ".sdat", in.fill.NIR)
NIR.filled <- raster(out.filled.NIR)
gc()

#### Band 6 flood fill
swir1=raster("im_B6.tif")

# estimating background (land) Band 5 ref
backg_B6=quantile(mask(swir1,idlnd,maskvalue=1),l_pt,na.rm=T);
swir1[mask[]==0]=backg_B6
rm(idlnd, mask); gc()

# fill in regional minimum Band 6 ref
in.fill.SWIR1 <- gsub(".grd$", ".sgrd", rasterTmpFile())
writeRaster(swir1,in.fill.SWIR1,format="SAGA",overwrite=TRUE,datatype = "FLT4S")
rsaga.fill.sinks(in.fill.SWIR1, in.fill.SWIR1, method = "xxl.wang.liu.2006", cores=1)
out.filled.SWIR1 <- gsub(".sgrd$", ".sdat", in.fill.SWIR1)
SWIR1.filled <- raster(out.filled.SWIR1)   
gc()

#### Compute shadow probability
Shadow[] <- (min(NIR.filled - nir, SWIR1.filled - swir1)>200)[]

# Remove extra files/objects
file.remove(in.fill.NIR, in.fill.SWIR1, out.filled.NIR, out.filled.SWIR1,  
            gsub(".sgrd$", ".mgrd", in.fill.NIR), gsub(".sgrd$", ".mgrd", in.fill.SWIR1))

# Release memory
rm(nir,swir1,NIR.filled,SWIR1.filled); gc()

cat("Read in potential mask", "\n")

# solar elevation angle
sun_ele=90-zen
sun_ele_rad=sun_ele*3.141593/180  # convert to radians

# solar azimuth anngle
sun_tazi=azi-90
sun_tazi_rad=sun_tazi*3.141593/180

# Next three lines unnecessary if using existing raster as mask? - GC
sub_size=resolu[1]
win_height=ijdim_ref[1]
win_width=ijdim_ref[2]

# potential cloud & shadow layer
cloud_test <- shadow_test <- raster(Cloud)
values(cloud_test) <- values(shadow_test) <- 0

# matched cloud & shadow layer
shadow_cal <- cloud_cal <- raster(Cloud)
values(shadow_cal) <- values(cloud_cal) <- 0

# cloud relative height (m) - RASTER
cloud_height <- raster(Cloud)
values(cloud_height) <- 0

# boundary layer
boundary_test <- raster(Cloud)
values(boundary_test) <- 0

# final cloud, shadow and snow mask
# get potential mask values
cs_final <- raster(Cloud)
values(cs_final) <- 0

# plshadow layer
shadow_test[Shadow[]==1]=1

# empty memory
rm(Shadow); gc()

# boundary layer
boundary_test[!is.na(Cloud[])]=1 

# plcloud layer
cloud_test[Cloud[]==1]=1

# empty memory
rm(Cloud); gc()

# revised percent of cloud on the scene after plcloud
revised_ptm=sum(cloud_test[])/sum(boundary_test[])

# no t test  => more than 99.9 % clouds and partly cloud over land
# => no match => rest are definite shadows
# cloud covers more than 90% of the scene
# => no match => rest are definite shadows

if ((ptm<=0.1)|(revised_ptm>=0.90)){
  cat("Skip cloud & cloud shadow matching because high cloud cover","\n")
  cloud_cal[cloud_test==1]=1
  shadow_cal[cloud_test==0]=1
  similar_num=-1
  height_num=-1
}  ### Add else statement!!!

cat("Cloud and cloud shadow matching ...","\n")

# define constants
Tsimilar=0.30

# threshold for matching buffering
Tbuffer=0.95

# max similarity threshold
max_similar = 0.95

# minimum matched cloud object (pixels)
num_cldoj=3

# number of inward pixes (90m) for cloud base temperature
num_pix=3

# enviromental lapse rate 6.5 degrees/km
# dry adiabatic lapse rate 9.8 degrees/km
rate_elapse=6.5  # degrees/km
rate_dlapse=9.8  # degrees/km

cat("Set cloud similarity = %.3","\n",Tsimilar)
cat("Set cloud buffer = %.3","\n",Tbuffer)
cat("Shadow match for cloud object >= %d pixels","\n",num_cldoj)

# move 2 pixel at a time
i_step=2*sub_size*tan(sun_ele_rad) 

##################################################################
# get moving direction
RowCol <- rowColFromCell(boundary_test, which(boundary_test[]==1))
y_ul <- max(RowCol[,2])
x_ul <- RowCol[min(which(RowCol[,2]==y_ul)),1]
y_lr <- min(RowCol[,2])
x_lr <- RowCol[max(which(RowCol[,2]==y_lr)),1]
x_ll <- min(RowCol[,1])
y_ll <- RowCol[min(which(RowCol[,1]==x_ll)),2]
x_ur <- max(RowCol[,1])
y_ur <- RowCol[max(which(RowCol[,1]==x_ur)),2]

# get view angle geometry
viewgeo <- function(x_ul,y_ul,x_ur,y_ur,x_ll,y_ll,x_lr,y_lr){
  x_u=(x_ul+x_ur)/2
  x_l=(x_ll+x_lr)/2
  y_u=(y_ul+y_ur)/2
  y_l=(y_ll+y_lr)/2  

  K_ulr=(y_ul-y_ur)/(x_ul-x_ur)  # get k of the upper left and right points
  K_llr=(y_ll-y_lr)/(x_ll-x_lr)  # get k of the lower left and right points
  K_aver=(K_ulr+K_llr)/2
  omiga_par=atan(K_aver)         # get the angle of parallel lines k (in pi)
  
  # AX(j)+BY(i)+C=0
  A=y_u-y_l
  B=x_l-x_u
  C=y_l*x_u-x_l*y_u
  
  omiga_per=atan(B/A)   # get the angle which is perpendicular to the trace line
  
  out <- list(A=A,B=B,C=C,omiga_par=omiga_par,omiga_per=omiga_per)
  out <- lapply(out,FUN=as.numeric)   # To remove attribute names
  return(out)
}

VIEWGEO <- viewgeo(x_ul,y_ul,x_ur,y_ur,x_ll,y_ll,x_lr,y_lr)
##################################################################

# Segmentate each cloud
cat("Cloud segmentation & matching","\n")
segm_cloud=clump(x=cloud_test,directions=8,gaps=FALSE)  # Sometimes still has gaps

# frecuency table of pixels by cloud object
area = tabulate(segm_cloud[])  # faster than freq/table

# filter out cloud object < than num_cldoj pixels
keep_idx = which(area >= num_cldoj)
segm_cloud[] = replace(segm_cloud[],!(segm_cloud[]%in%keep_idx),NA)

# renumber cloud objects
segm_cloud = reclassify(segm_cloud,rcl=cbind(keep_idx,1:length(keep_idx)))

# updated frecuency table of pixels by cloud object
obj_num = tabulate(segm_cloud[])
num=length(obj_num)
area_final = cbind(1:num,obj_num)

# Get the x,y of each cloud
# Matrix used in recording the x,y
xys = cbind(segm_cloud[which(!is.na(segm_cloud[]))], rowColFromCell(segm_cloud, which(!is.na(segm_cloud[]))))

# Use iteration to get the optimal move distance
# Calulate the moving cloud shadow
height_num=rep(0,num)   # cloud relative height (m)
similar_num=rep(0,num)

mat_truecloud <- function(x,y,h,A,B,C,omiga_par,omiga_per){
  # imput "x",j col
  # imput "y",i row
  # imput cloud height "h"
  H=705000   # average Landsat 4,5,&7 height (m)
  dist=(A*x+B*y+C)/((A^2+B^2)^(0.5))  # from the cetral perpendicular (unit: pixel)
  dist_par=dist/cos(omiga_per-omiga_par)
  dist_move=dist_par*h/H   # cloud move distance (m)
  delt_x=dist_move*cos(omiga_par)
  delt_y=dist_move*sin(omiga_par)
  x_new=x+delt_x  # new x, j
  y_new=y+delt_y  # new y, i  
  return(cbind(x_new,y_new))
}

# cloud shadow match similarity (m)
im_B11 <- raster("im_B11.tif") # Read in temperature band

for (cloud_type in 1:num){

  cat("Matching Cloud #",cloud_type,"of",num,"\n")
  
  # moving cloud xys
  XY_type=matrix(0,nrow=obj_num[cloud_type],ncol=2)

  # record the max threshold moving cloud xys
  tmp_XY_type=matrix(0,nrow=obj_num[cloud_type],ncol=2)
    
  # corrected for view angle xys
  tmp_xys=matrix(0,nrow=obj_num[cloud_type],ncol=2)

  # record the original xys
  orin_xys=matrix(0,nrow=obj_num[cloud_type],ncol=2)

  # record the original xys
  orin_xys[]=xys[which(xys[,1]==cloud_type),2:3]

  # record this original ids
  orin_cid=cellFromRowCol(im_B11,rownr=orin_xys[,1],colnr=orin_xys[,2])

  # Temperature of the cloud object
  temp_obj=im_B11[orin_cid]      # Extract temperature
  
  # the base temperature for cloud
  # assume object is round r_obj is radium of object
  r_obj=sqrt(obj_num[cloud_type]/3.141593)
  
  #  number of inward pixes for correct temperature
  #  num_pix=8     --- actually num_pix=3 above in constants (GC)
  pct_obj=(r_obj-num_pix)^2/r_obj^2
  pct_obj=min(pct_obj,1)  #  pct of edge pixel should be less than 1
  t_obj=quantile(temp_obj,pct_obj)
  
  #  put the edge of the cloud the same value as t_obj
  temp_obj[temp_obj>t_obj]=t_obj
  
  #  wet adiabatic lapse rate 6.5 degrees/km
  #  dry adiabatic lapse rate 9.8 degrees/km
  #  rate_wlapse=6.5   # degrees/km
  #  rate_dlapse=9.8   # degrees/km
  
  Max_cl_height=12000  # Max cloud base height (m)
  Min_cl_height=200    # Min cloud base height (m)
  
  # refine cloud height range (m)
  Min_cl_height=max(Min_cl_height,10*(Tland[1]-400-t_obj)/rate_dlapse)
  Max_cl_height=min(Max_cl_height,10*(Tland[2]+400-t_obj))
  
  # initialize height and similarity info
  record_h=0
  record_thresh=0

  for (base_h in seq(from=Min_cl_height,to=Max_cl_height,by=i_step)){
    
    # iterate in height (m)
    # Get the true postion of the cloud
    # calculate cloud DEM with initial base height
    h=(10*(t_obj-temp_obj)/rate_elapse+base_h)
  
    tmp_xys=mat_truecloud(x=orin_xys[,2],y=orin_xys[,1],h = h,A = VIEWGEO$A,B = VIEWGEO$B,
                          C = VIEWGEO$C,omiga_par = VIEWGEO$omiga_par,omiga_per = VIEWGEO$omiga_per)
  
    # shadow moved distance (pixel)
    # i_xy=h*cos(sun_tazi_rad)/(sub_size*tan(sun_ele_rad))
    i_xy=h/(sub_size*tan(sun_ele_rad))
  
    if (azi < 180){
      XY_type[,2]=round(tmp_xys[,1]-i_xy*cos(sun_tazi_rad));  # X is for j,2
      XY_type[,1]=round(tmp_xys[,2]-i_xy*sin(sun_tazi_rad));  # Y is for i,1
    } else{
      XY_type[,2]=round(tmp_xys[,1]+i_xy*cos(sun_tazi_rad));  # X is for j,2
      XY_type[,1]=round(tmp_xys[,2]+i_xy*sin(sun_tazi_rad));  # Y is for i,1
    }
  
    tmp_j=XY_type[,2]  # col
    tmp_i=XY_type[,1]  # row
  
    # the id that is out of the image
    out_id=(tmp_i<1)|(tmp_i>win_height)|(tmp_j<1)|(tmp_j>win_width)
    out_all=sum(out_id)
  
    tmp_ii=tmp_i[out_id==0]
    tmp_jj=tmp_j[out_id==0]

    tmp_id=cellFromRowCol(im_B11,tmp_ii,tmp_jj)
    
    # the id that is matched (exclude original cloud)
    match_id=(boundary_test[tmp_id]==0)|                            # Are points outside boundary? - GC
      (((segm_cloud[tmp_id]!=cloud_type)|is.na(segm_cloud[tmp_id])) # If not, points must be clear (NA) or another cloud - GC
      &((cloud_test[tmp_id]>0)|(shadow_test[tmp_id]==1)))           # If so, points must be cloud or shadow to match - GC
    matched_all=sum(match_id)+out_all
  
    # the id that is the total pixel (exclude original cloud)
    total_id=(segm_cloud[tmp_id]!=cloud_type)|is.na(segm_cloud[tmp_id])
    total_all=sum(total_id)+out_all
  
    thresh_match=matched_all/total_all
    
    if ((thresh_match >= Tbuffer*record_thresh)&(base_h < Max_cl_height-i_step)&(record_thresh < max_similar)){
      if (thresh_match > record_thresh){
        record_thresh=thresh_match;
        record_h=h;
      } else if (record_thresh > Tsimilar){
        similar_num[cloud_type]=record_thresh;
        i_vir=record_h/(sub_size*tan(sun_ele_rad));
        # height_num=record_h;
        
        if (azi < 180){
          tmp_XY_type[,2]=round(tmp_xys[,1]-i_vir*cos(sun_tazi_rad));  # X is for col j,2
          tmp_XY_type[,1]=round(tmp_xys[,2]-i_vir*sin(sun_tazi_rad));  # Y is for row i,1
        } else{
          tmp_XY_type[,2]=round(tmp_xys[,1]+i_vir*cos(sun_tazi_rad));  # X is for col j,2
          tmp_XY_type[,1]=round(tmp_xys[,2]+i_vir*sin(sun_tazi_rad));  # Y is for row i,1
        }
  
        tmp_scol=tmp_XY_type[,2];
        tmp_srow=tmp_XY_type[,1];
        
        # put data within range
        tmp_srow[tmp_srow<1] <- 1;
        tmp_srow[tmp_srow>win_height] <- win_height;
        tmp_scol[tmp_scol<1] <- 1;
        tmp_scol[tmp_scol>win_width] <- win_width;
        
        tmp_sid=cellFromRowCol(im_B11,tmp_srow,tmp_scol);
        
        # give shadow_cal=1
        shadow_cal[tmp_sid] <- 1;
        # record matched cloud
        # cloud_cal(orin_cid)=1;
        # cloud_height(orin_cid)=record_h;
      } else{
        record_thresh=0;
      }
    }
  }
}

rm(im_B11); gc()

### dilate each cloud and shadow object by 3 and 3 pixel outward in 8 connect directions
cldpix=3   # number of pixels to be dilated for cloud
sdpix=3    # number of pixels to be dilated for shadow
snpix=3    # number of pixels to be dilated for snow
    
cat("Dilate (3) pixels for cloud, shadow, and snow objects","\n")

SEc=matrix(1,nrow=cldpix,ncol=cldpix)
SEs=matrix(1,nrow=sdpix,ncol=sdpix)
SEw=matrix(1,nrow=snpix,ncol=snpix)
      
# dialte shadow first
shadow_cal=focal(shadow_cal,w=SEs)
shadow_cal[]=1*(shadow_cal[]>0)

# find shadow within plshadow
# shadow_cal[shadow_test!=1] <- 0
# dilate shadow again with the more accurate cloud shadow
# shadow_cal=focal(shadow_cal,w=SEs)
# shadow_cal[]=1*(shadow_cal[]>0)

cloud_cal = segm_cloud
cloud_cal[] <- 1*(cloud_cal[]>0)
cloud_cal=focal(cloud_cal,w=SEc)
cloud_cal[]=1*(cloud_cal[]>0)
        
# dilate snow by snpix pixels in 8 connect directions (Zhe 09/12/2012)
Snow=raster("SnowMask.tif")
Snow=focal(Snow,w=SEw)
Snow[]=1*(Snow[]>0)

# mask from plcloud
cs_final[WT[]==1]=1
rm(WT); gc()

# step 1 snow or unknow
cs_final[Snow[]==1]=3  # snow

# step 2 shadow above snow and everyting
cs_final[shadow_cal[]==1]=2  # shadow
        
# step 3 cloud above all
cs_final[cloud_cal[]==1]=4  # cloud
        
cs_final[boundary_test[]==0]=NA

# record clear pixel percent;
tmpcs = (cs_final <= 1)
clr_pct = 100*sum(tmpcs[],na.rm = T)/sum(boundary_test[])

writeRaster(cs_final, filename="FinalMask.tif", format="GTiff", overwrite=TRUE)   # write that raster to file

cat("Tile",TILE,"Percent Clear Pixels ---",clr_pct,"%","\n")

# Remove cloud pixels from each Landsat Band
for(i in BANDS[BANDS!=9]){
  raster <- raster(paste0("im","_B",i,".tif"))
  raster[which((cs_final[]==2)|(cs_final[]==4))] <- NA
  writeRaster(raster, filename=paste0("im","_B",i,".tif"), format="GTiff", overwrite=TRUE)   # write that raster to file
}

rm(list= ls()[!(ls() %in% c('BANDS','baseDir','OutFolder','ScriptFolder','DEMFileName','imageDirArray','tmpDIR',
                            's_zen','azi','ScriptFolder','TILE','zen','gcr'))]); gc()

